"""Module with the `ContextData` class.

This module contains the `ContextData` class that can be used to specify the
context-variables in which a `Context` object should operate.
"""
from my_model.user_scoped_models import User  # type: ignore
from sqlalchemy.future import Engine
from sqlmodel import Session


class ContextData:
    """Class with ContextData.

    Dataclass that contains the data for a `Context`-object. The properties
    defined in this class define the context that is being used.

    Attributes:
        user: a user object representing this context.
        db_session: a SQLalchemy session that can be used.
    """

    def __init__(self, database_engine: Engine, user: User) -> None:
        """Create the ContextData object.

        The initiator sets the values for the ContextData and creates a
        SQLalchmey Session.

        Args:
            database_engine: a SQLalchemy Engine to bind the Session to.
            user: a User to bind the Context to
        """
        self.user: User = user
        self.db_session = Session(database_engine, expire_on_commit=False)

    def close_session(self) -> None:
        """Commit and closes the database sessions.

        Method to commit the changes in the session and close the session. This
        should be done when the Context is ready with this ContextData object.
        """
        if self.db_session:
            self.db_session.commit()
            self.db_session.close()



"""Module with `my-data` exceptions.

Contains exceptions for `my-data`.
"""


class MyDataException(Exception):
    """Base exception for My-Data-exceptions."""


class DatabaseNotConfiguredException(MyDataException):
    """Raised when the database is not configured."""


class DatabaseConnectionException(MyDataException):
    """Error that happends when the database string is not correct."""


class BaseClassCallException(MyDataException):
    """The Base class is used for a method that should be overriden."""


class WrongDataManipulatorException(MyDataException):
    """Wrong DataManipulator for specified model."""


class PermissionDeniedException(MyDataException):
    """User tries to do something he is not allowed to do."""


class UnknownUserAccountException(MyDataException):
    """Service user tries to retrieve a non-existing user account."""



"""Module with Creators.

This module contains the creator classes. These classes are used to create data
in the database. The ResourceManager uses these classes.
"""
from typing import TypeVar

from my_model.user_scoped_models import (User, UserRole,  # type: ignore
                                         UserScopedModel)

from .data_manipulator import DataManipulator
from .exceptions import (BaseClassCallException, PermissionDeniedException,
                         WrongDataManipulatorException)

T = TypeVar('T')


class Creator(DataManipulator):
    """Baseclass for Creators.

    The baseclass for creators. The sub creators use this class to make sure
    creators have the same interface.
    """

    def is_authorized(self) -> bool:
        """Authorize the creation of this data.

        Method that checks if the current context is allowd to create this type
        of model.

        Raises:
            BaseClassCallException: BaseClass method is used.
        """
        raise BaseClassCallException('Method not implemented in baseclass')

    def create(self, models: list[T] | T) -> list[T]:
        """Create data.

        The method to create data in the database.

        Args:
            models: the models to create.

        Raises:
            PermissionDeniedException: when the user is not authorized to
                create this resource.

        Returns:
            A list with the created data models.
        """
        if self._context_data.user.role not in (UserRole.USER, UserRole.ROOT):
            raise PermissionDeniedException(
                'User must be a normal user or a root user to create data.')

        if not self.is_authorized():
            raise PermissionDeniedException(
                'Not allowed to create this kind of object within the ' +
                'set context.')
        return self._add_models_to_session(models)


class UserScopedCreator(Creator):
    """Creator for UserScoped models.

    This creator should be used for UserScoped models, like Tags and APItokens.
    """

    def is_authorized(self) -> bool:
        """Check if this user can create tags.

        This method checks if the given databasemodel is, in fact, a UserScoped
        model. If it isn't, it raises a WrongDataManipulatorException. If it
        is, is returns True, indicating that this user can create this type of
        resource.

        Raises:
            WrongDataManipulatorException: when the database_model for this
                instance is not correct.

        Returns:
            True when the user can create these types of objects.
        """
        if not issubclass(self._database_model, UserScopedModel):
            raise WrongDataManipulatorException(
                f'The model "{self._database_model}" is not a UserScopedModel')
        return self._context_data.user.role in (UserRole.ROOT, UserRole.USER)

    def create(self, models: list[T] | T) -> list[T]:
        """Create the UserScoped data.

        We override this method from the superclass because we have to make
        sure the `user_id` is set to the correct value first. If this field is
        already set to a wrong user_id, we raise an exception.

        Args:
            models: the models to create.

        Raises:
            PermissionDeniedException: when the model has a user_id set that is
            different then the current user_id in the context.

        Returns:
            A list with the created data models.
        """
        # Make sure the `models` are always a list
        models = self._convert_model_to_list(models)

        # Add the `user_id` attribute or raise an error when it is already set
        # to a wrong value
        for model in models:
            user_id = getattr(model, 'user_id', None)
            if user_id is not None and user_id != self._context_data.user.id:
                raise PermissionDeniedException(
                    'This user is not allowed to create this resource')
            setattr(model, 'user_id', self._context_data.user.id)

        return super().create(models)


class UserCreator(Creator):
    """Creator for Users.

    This creator should be used to create Users.
    """

    def is_authorized(self) -> bool:
        """Check if this user can create users.

        Checks where the given datamodel is, in fact a User. If it isn't, it
        raises a WrongDataManipulatorException. If it is, it checks the user in
        the context; if there is a user set, it should be a ROOT user. If there
        is no user set, the request is authorized.

        Raises:
            WrongDataManipulatorException: when the database_model for this
                instance is not correct.

        Returns:
            True is this user can create Users and False if this user can't.
        """
        if self._database_model is not User:
            raise WrongDataManipulatorException(
                f'The model "{self._database_model}" is not a UserScopedModel')
        return (self._context_data.user is None or
                self._context_data.user.role == UserRole.ROOT)



"""Module with the ResourceManager class.

Contains the ResourceManager class that is used by a Context to create a
ResourceManager for specific resources.
"""
from typing import Generic, Type, TypeVar

from sqlalchemy.future import Engine
from sqlalchemy.sql.elements import ColumnElement

from .context_data import ContextData
from .creators import Creator, UserScopedCreator
from .deleters import Deleter, UserScopedDeleter
from .retrievers import Retriever, UserScopedRetriever
from .updaters import Updater, UserScopedUpdater

T = TypeVar('T')


class ResourceManager(Generic[T]):
    """Manager for resources in the database.

    The ResourceManager class is used to manage resources in the database in a
    CRUD way. This means that this class gets a Creator, Retriever, Updater and
    Deleter class assigned to create, retrieve, update and delete resources.

    Attributes:
        _database_model: the SQLmodel model used by this ResourceManager.
        _database_engine: the SQLalchemy engine to use.
        _context_data: specifies in what context to execute the methods.
        retriever: a instance of a subclass of Retriever. This retrieves the
            actual data.
        creator: a instance of a subclass of Creator. This creates the actual
            data.
        updater: a instance of a subclass of Updater. This updates the actual
            data.
        deleter: a instance of a subclass of Deleter. This deletes the actual
            data.
    """

    def __init__(self,
                 database_model: Type[T],
                 database_engine: Engine,
                 context_data: ContextData | None = None,
                 creator: Type = UserScopedCreator,
                 retriever: Type = UserScopedRetriever,
                 updater: Type = UserScopedUpdater,
                 deleter: Type = UserScopedDeleter) -> None:
        """Set attributes for the object.

        The initiator sets the attributes for the object.

        Args:
            database_model: the SQLmodel model used by this ResourceManager.
            database_engine: the SQLalchemy engine to use.
            context_data: specifies in what context to execute the methods.
            creator: the class for the Creator.
            retriever: the class for the Retriever.
            updater: the class for the Updater.
            deleter: the class for the Deleter.
        """
        self._database_model: Type = database_model
        self._database_engine: Engine = database_engine
        self._context_data: ContextData | None = context_data

        self.retriever: Retriever = retriever(
            database_model=database_model,
            database_engine=database_engine,
            context_data=context_data
        )
        self.creator: Creator = creator(
            database_model=database_model,
            database_engine=database_engine,
            context_data=context_data
        )
        self.updater: Updater = updater(
            database_model=database_model,
            database_engine=database_engine,
            context_data=context_data
        )
        self.deleter: Deleter = deleter(
            database_model=database_model,
            database_engine=database_engine,
            context_data=context_data
        )

    def create(self, models: list[T] | T) -> list[T]:
        """Create resources.

        Creates one or multiple resources. It uses the defined creator to
        create the resource in the correct way.

        Args:
            models: the model or models to add.

        Returns:
            Model: the new model.
        """
        return self.creator.create(models)

    def retrieve(
            self,
            flt: list[ColumnElement] | ColumnElement | None = None) -> list[T]:
        """Retrieve resources for the specified object.

        Returns a list of resources for the specified model. It does this using
        the defined retriever to make sure it partains to the specified
        ContextData-object.

        Args:
            flt: SQLModel type filters to filter this resource.

        Returns:
            list[Model]: a list with the retrieved resources in models defined
                in the `my-models` package.
        """
        # Get all DB objects from the database
        return self.retriever.retrieve(flt=flt)

    def update(self, models: list[T] | T) -> list[T]:
        """Update resources.

        Updates one ore more resources. It uses the defined Updater to update
        the resource in the correct way.

        Args:
            models: the model or models to update.

        Returns:
            Model: the updated model.
        """
        return self.updater.update(models)

    def delete(self, models: list[T] | T) -> None:
        """Delete resources.

        Deletes one or more resources. It uses the defined Deleter to delete
        the resource in the correct way.

        Args:
            models: the model or models to update.
        """
        self.deleter.delete(models)



"""Module with Updaters.

This module contains the updator classes. These classes are used to update data
in the database. The ResourceManager uses these classes.
"""

from typing import TypeVar

from my_model.user_scoped_models import User, UserRole  # type: ignore

from my_data.exceptions import (PermissionDeniedException,
                                WrongDataManipulatorException)

from .data_manipulator import DataManipulator

T = TypeVar('T')


class Updater(DataManipulator):
    """Baseclass for Updaters.

    The baseclass for updaters. The sub updaters use this class to make sure
    updaters have the same interface.
    """

    def update(self, models: list[T] | T) -> list[T]:
        """Update data.

        The method to update data in the database.

        Args:
            models: the models to update.

        Raises:
            PermissionDeniedException: when this user type is not allowed to
                update data this way.

        Returns:
            A list with the updated data models.
        """
        if self._context_data.user.role not in (UserRole.USER, UserRole.ROOT):
            raise PermissionDeniedException(
                'User must be a normal user or a root user to update data.')
        return self._add_models_to_session(models)


class UserScopedUpdater(Updater):
    """Updater for UserScoped models.

    This updater should be used for UserScoped models, like Tags and APItokens.
    """

    def update(self, models: list[T] | T) -> list[T]:
        """Update the UserScoped data.

        We override this method from the superclass because we have to make
        sure the `user_id` is set to the correct value first. If this field is
        set to a wrong user_id, we raise an exception.

        Args:
            models: the models to update.

        Returns:
            A list with the created data models.
        """
        models = self._validate_user_scoped_models(models)
        return super().update(models)


class UserUpdater(Updater):
    """Updater for Users.

    This updaters should be used to update Users.
    """

    def update(self, models: list[T] | T) -> list[T]:
        """Update the User data.

        We override this method from the superclass because we have to make
        sure the model is a User model and that the user in the context is
        allowed to update this user. A root user can update ALL users, but a
        normal user can only update his own user.

        Args:
            models: the models to update.

        Raises:
            WrongDataManipulatorException: when the model in the instance is
                not a User.
            PermissionDeniedException: when the model is not the same model as
                set in the instance or when the model has a id set that is
                different then the current user id in the context.

        Returns:
            A list with the created data models.
        """
        if self._database_model is not User:
            raise WrongDataManipulatorException(
                f'The model "{self._database_model}" is not a User')

        # Make sure the `models` are always a list
        models = self._convert_model_to_list(models)

        # Check for all models are a User model and if the `id` field is the
        # same as the current user if this user is a USER user.
        for model in models:
            if not isinstance(model, User):
                raise PermissionDeniedException(
                    f'Expected "{self._database_model}", got "{type(model)}".')

            if self._context_data.user.role == UserRole.USER:
                if model.id != self._context_data.user.id:
                    raise PermissionDeniedException(
                        'User is not allowed to edit this user.')

        return super().update(models)



"""Module for the Context class.

This module contains the Context class.
"""
from types import TracebackType

from my_model.user_scoped_models import (APIClient, APIToken,  # type: ignore
                                         Tag, User, UserRole, UserSetting)
from sqlalchemy.future import Engine
from sqlmodel import select

from .context_data import ContextData
from .creators import UserCreator, UserScopedCreator
from .deleters import UserDeleter, UserScopedDeleter
from .exceptions import PermissionDeniedException, UnknownUserAccountException
from .resource_manager import ResourceManager
from .retrievers import UserRetriever, UserScopedRetriever
from .updaters import UserScopedUpdater, UserUpdater


class Context:
    """Context to work in.

    The Context class is used to create objects to interact with the database
    in a context-based way. A Context uses a ContextData object, which contains
    the data for the context (like a User to work with). The Context makes sure
    that every manipulation of the database is done with the permissions of the
    Context.

    Attributes:
        database_engine: the SQLalchemy engine to use.
        _context_data: specifies in what context to use Context.
    """

    def __init__(self,
                 database_engine: Engine,
                 context_data: ContextData) -> None:
        """Set the default values and create the needed DataManipulators.

        The initializer set the values and creates the needed DataManipulator
        objects. These objects are the objects that are used to manipulate the
        data in the database, like users and tags.

        Args:
            database_engine: a database engine to work with.
            context_data: the context data for this context.
        """
        self.database_engine = database_engine
        self._context_data = context_data

        # Exposure of Resource Managers to manage specific resources in the
        # data model.
        self.users = ResourceManager(
            database_model=User,
            database_engine=database_engine,
            context_data=self._context_data,
            creator=UserCreator,
            retriever=UserRetriever,
            updater=UserUpdater,
            deleter=UserDeleter)
        self.tags = ResourceManager(
            database_model=Tag,
            database_engine=database_engine,
            context_data=self._context_data,
            creator=UserScopedCreator,
            retriever=UserScopedRetriever,
            updater=UserScopedUpdater,
            deleter=UserScopedDeleter)
        self.api_clients = ResourceManager(
            database_model=APIClient,
            database_engine=database_engine,
            context_data=self._context_data,
            creator=UserScopedCreator,
            retriever=UserScopedRetriever,
            updater=UserScopedUpdater,
            deleter=UserScopedDeleter)
        self.api_tokens = ResourceManager(
            database_model=APIToken,
            database_engine=database_engine,
            context_data=self._context_data,
            creator=UserScopedCreator,
            retriever=UserScopedRetriever,
            updater=UserScopedUpdater,
            deleter=UserScopedDeleter)
        self.user_settings = ResourceManager(
            database_model=UserSetting,
            database_engine=database_engine,
            context_data=self._context_data,
            creator=UserScopedCreator,
            retriever=UserScopedRetriever,
            updater=UserScopedUpdater,
            deleter=UserScopedDeleter)

    def __enter__(self) -> 'Context':
        """Start a Python context manager.

        The start of a Context Manager. Should be used with the Python `with`
        statement.

        Returns:
            This own class.
        """
        return self

    def __exit__(self,
                 exception_type: BaseException | None,
                 exception_value: BaseException | None,
                 traceback: TracebackType | None) -> bool:
        """Exit of a Python context manager.

        The end of the context manager. Checks if there are any unhandled
        execeptions and returns True if there aren't.

        Args:
            exception_type: the type of exception that happened.
            exception_value: the value for the exception.
            traceback: a traceback for the exception.

        Returns:
            False if there are unhandled exceptions, True if there are none.
        """
        self._context_data.close_session()
        return exception_type is None

    def get_user_account_by_username(self, username: str) -> User:
        """Get a User account using a service account.

        Retrieves a User account for a user by searching for a specific
        username. This can only be done by a service user.

        Args:
            username: the username for the user.

        Raises:
            PermissionDeniedException: user in the context is not a Service
                user.
            UnknownUserAccountException: the user is not found.

        Returns:
            A User object for the correct user.
        """
        if self._context_data.user.role != UserRole.SERVICE:
            raise PermissionDeniedException('Should be ran as a service user')

        sql_query = select(User).where(User.username == username)
        if self._context_data.db_session:
            user_object = self._context_data.db_session.exec(sql_query).all()
        if len(user_object) == 1:
            return user_object[0]
        raise UnknownUserAccountException(
            f'User with username "{username}" is not found.')

    def get_user_account_by_api_token(self, api_token: str) -> User:
        """Get a User account using via a API token.

        Retrieves a User account for a user by searching for a specific API
        token. This can only be done by a service user.

        Args:
            api_token: the API token for the user.

        Raises:
            PermissionDeniedException: user in the context is not a Service
                user.
            UnknownUserAccountException: the user is not found.

        Returns:
            A User object for the correct user.
        """
        if self._context_data.user.role != UserRole.SERVICE:
            raise PermissionDeniedException('Should be ran as a service user')

        sql_query = select(APIToken).where(APIToken.token == api_token)
        api_tokens = self._context_data.db_session.exec(sql_query).all()
        if len(api_tokens) == 1:
            return api_tokens[0].user
        raise UnknownUserAccountException(
            f'Token "{api_token}" is not found.')



"""Module with the main class `MyData`.

This module contains the class `MyData`, which is the most important class for
the complete project.
"""

from typing import Any

from my_model.user_scoped_models import (APIClient, APIToken,  # type: ignore
                                         Tag, User, UserRole, UserSetting)
from sqlalchemy.exc import OperationalError
from sqlalchemy.future import Engine
from sqlmodel import Session, SQLModel, and_, create_engine, select

from .context import Context
from .context_data import ContextData
from .exceptions import (DatabaseConnectionException,
                         DatabaseNotConfiguredException,
                         PermissionDeniedException)


class MyData:
    """Main class for the library.

    This class creates the database connection and has a method to expose a
    context.

    Attributes:
        database_engine: the SQLmodel Engine.
        _database_str: the database connection string.
        _database_args: the arguments for the SQLalchemy connection.
    """

    def __init__(self) -> None:
        """Set default values.

        The initiator sets all values to `None`. We can later override these
        values with the `configure` method.
        """
        self.database_engine: Engine | None = None
        self._database_str: str | None = None
        self._database_args: dict[str, Any] | None = None

    def configure(self,
                  db_connection_str: str,
                  database_args: dict[str, Any] | None = None) -> None:
        """Set the database configuration.

        The database configuration is used when the database connection has to
        be made.

        Args:
            db_connection_str: the database connection string for the
                SQLmodel model.
            database_args: a dict with extra configuration for SQLmodel.
        """
        self._database_str = db_connection_str
        self._database_args = database_args

    def create_engine(self, force: bool = False) -> None:
        """Create the database connection.

        Creates the database connection with the given configuration. If the
        configuration is not set, it will raise an exception.

        Args:
            force: if set to True, the database connection will be created even
                if the database connection is already created.

        Raises:
            DatabaseNotConfiguredException: database not configured.
            DatabaseConnectionException: error while connecting to the
                database.
        """
        # Stop if there is already a database engine
        if self.database_engine is not None and not force:
            return

        # Check if the database connection is set
        if self._database_str is None:
            raise DatabaseNotConfiguredException(
                'Database is not configured yet')

        # Connect to the database
        database_args: dict[str, Any] = {
            'url': self._database_str
        }
        if self._database_args:
            database_args.update(self._database_args)

        # Create the database engine
        try:
            self.database_engine = create_engine(**database_args)
        except OperationalError as sa_error:
            raise DatabaseConnectionException(
                'Couldn\'t connect to database') from sa_error

    def create_db_tables(self, drop_tables: bool = False) -> None:
        """Create the defined models as tables.

        Method to create all tables that are defined in models.

        Args:
            drop_tables: determines if tables should be dropped prior to
                creating them. SQLalchemy will only drop the tables it
                knows about.
        """
        self.create_engine()

        if drop_tables:
            SQLModel.metadata.drop_all(self.database_engine)
        SQLModel.metadata.create_all(self.database_engine)

    def get_context(self, user: User) -> Context:
        """Get a Context object for this database.

        Method to create a Context object and return it. The returned Context
        can be used in context manager right away, or as a `normal` object.

        Args:
            user: the user for the context. Is used in the ContextData object
                that is created for the Context object.

        Raises:
            DatabaseNotConfiguredException: method is called before configuring
                the database.

        Returns:
            The created Context object.
        """
        self.create_engine()

        if not self.database_engine:
            raise DatabaseNotConfiguredException(
                'Database is not configured yet')

        return Context(
            database_engine=self.database_engine,
            context_data=ContextData(
                database_engine=self.database_engine,
                user=user)
        )

    def get_context_for_service_user(
            self, username: str, password: str) -> Context:
        """Get a Context object for the database as a service user.

        Method to create a Context object for a service user. The context can
        be used to do the things a service user needs to do, like retrieving
        user objects or API token objects.

        Args:
            username: the username for the service user.
            password: the password for the service user.

        Raises:
            DatabaseNotConfiguredException: method is called before configuring
                the database.
            PermissionDeniedException: when the credentials are incorrect.

        Returns:
            The created Context object.
        """
        self.create_engine()

        if not self.database_engine:
            raise DatabaseNotConfiguredException(
                'Database is not configured yet')

        with Session(self.database_engine) as session:
            sql_query = select(User)
            sql_query = sql_query.where(
                and_(User.username == username,
                     User.role == UserRole.SERVICE))
            users = session.exec(sql_query).all()

            # Check the amount of users we got
            if len(users) != 1:
                raise PermissionDeniedException(
                    f'Service account "{username}" does not exist')

            # Check if the provided credentials are correct
            user = users[0]
            if not user.verify_credentials(username, password):
                raise PermissionDeniedException(
                    f'Password for Service account "{username}" is incorrect')

            return Context(
                database_engine=self.database_engine,
                context_data=ContextData(
                    database_engine=self.database_engine,
                    user=user)
            )

    def create_init_data(self) -> None:
        """Create initializer data.

        Method to create initialization data in the database. Creates data in
        the database that can be used to initialize a database or to create a
        test database. Warning: this method will erase the complete database!
        """
        self.create_db_tables(drop_tables=True)

        with Session(self.database_engine) as session:
            users: list[User] = [
                User(
                    id=1,
                    fullname='root',
                    username='root',
                    email='root@example.com',
                    role=UserRole.ROOT,
                    tags=[
                        Tag(title='root_tag_1'),
                        Tag(title='root_tag_2'),
                        Tag(title='root_tag_3')
                    ],
                    api_clients=[
                        APIClient(
                            app_name='root_api_client_1',
                            app_publisher='root_api_client_1_publisher'),
                        APIClient(
                            app_name='root_api_client_2',
                            app_publisher='root_api_client_2_publisher'),
                        APIClient(
                            app_name='root_api_client_3',
                            app_publisher='root_api_client_3_publisher')
                    ],
                    api_tokens=[
                        APIToken(title='root_api_token_1'),
                        APIToken(title='root_api_token_2'),
                        APIToken(title='root_api_token_3')
                    ],
                    user_settings=[
                        UserSetting(setting='root_test_setting_1',
                                    value='test_value_1'),
                        UserSetting(setting='root_test_setting_2',
                                    value='test_value_2'),
                        UserSetting(setting='root_test_setting_3',
                                    value='test_value_3')
                    ]
                ),
                User(
                    id=2,
                    fullname='Normal user 1',
                    username='normal.user.1',
                    email='normal_user_1@example.com',
                    role=UserRole.USER,
                    tags=[
                        Tag(title='normal_user_1_tag_1'),
                        Tag(title='normal_user_1_tag_2'),
                        Tag(title='normal_user_1_tag_3')
                    ],
                    api_clients=[
                        APIClient(
                            app_name='normal_user_1_api_client_1',
                            app_publisher='normal_user_1_api_client_1_' +
                            'publisher'),
                        APIClient(
                            app_name='normal_user_1_api_client_2',
                            app_publisher='normal_user_1_api_client_2_' +
                            'publisher'),
                        APIClient(
                            app_name='normal_user_1_api_client_3',
                            app_publisher='normal_user_1_api_client_3_' +
                            'publisher')
                    ],
                    api_tokens=[
                        APIToken(title='normal_user_1_api_token_1'),
                        APIToken(title='normal_user_1_api_token_2'),
                        APIToken(title='normal_user_1_api_token_3')
                    ],
                    user_settings=[
                        UserSetting(setting='normal_user_1_test_setting_1',
                                    value='test_value_1'),
                        UserSetting(setting='normal_user_1_test_setting_2',
                                    value='test_value_2'),
                        UserSetting(setting='normal_user_1_test_setting_3',
                                    value='test_value_3')
                    ]
                ),
                User(
                    id=3,
                    fullname='Normal user 2',
                    username='normal.user.2',
                    email='normal_user_2@example.com',
                    role=UserRole.USER,
                    tags=[
                        Tag(title='normal_user_2_tag_1'),
                        Tag(title='normal_user_2_tag_2'),
                        Tag(title='normal_user_2_tag_3')
                    ],
                    api_clients=[
                        APIClient(
                            app_name='normal_user_2_api_client_1',
                            app_publisher='normal_user_2_api_client_1_' +
                            'publisher'),
                        APIClient(
                            app_name='normal_user_2_api_client_2',
                            app_publisher='normal_user_2_api_client_2_' +
                            'publisher'),
                        APIClient(
                            app_name='normal_user_1_api_client_3',
                            app_publisher='normal_user_2_api_client_3_' +
                            'publisher')
                    ],
                    api_tokens=[
                        APIToken(title='normal_user_2_api_token_1',
                                 token='aRlIytpyz61JX2TvczLxJZUsRzk578pE'),
                        APIToken(title='normal_user_2_api_token_2'),
                        APIToken(title='normal_user_2_api_token_3')
                    ],
                    user_settings=[
                        UserSetting(setting='normal_user_2_test_setting_1',
                                    value='test_value_1'),
                        UserSetting(setting='normal_user_2_test_setting_2',
                                    value='test_value_2'),
                        UserSetting(setting='normal_user_2_test_setting_3',
                                    value='test_value_3')
                    ]
                ),
                User(
                    id=4,
                    fullname='Service User - for tests',
                    username='service.user',
                    email='service.user@example.com',
                    role=UserRole.SERVICE,
                )]
            users[0].set_password('root_pw')
            users[1].set_password('normal_user_1_pw')
            users[2].set_password('normal_user_2_pw')
            users[3].set_password('service_password')
            for user in users:
                session.add(user)
            session.commit()



"""Module with Deleters.

This module contains the deleter classes. These classes are used to delete data
from the database. The ResourceManager uses these classes.
"""
from typing import TypeVar

from my_model.user_scoped_models import User, UserRole  # type: ignore

from my_data.exceptions import (PermissionDeniedException,
                                WrongDataManipulatorException)

from .data_manipulator import DataManipulator

T = TypeVar('T')


class Deleter(DataManipulator):
    """Baseclass for Deleters.

    The baseclass for deleters. The sub deleters use this class to make sure
    deleters have the same interface.
    """

    def delete(self, models: list[T] | T) -> None:
        """Delete data.

        The method to delete data from the database.

        Args:
            models: the models to delete.

        Raises:
            PermissionDeniedException: when this user type is not allowed to
                delete data this way.
        """
        if self._context_data.user.role not in (UserRole.USER, UserRole.ROOT):
            raise PermissionDeniedException(
                'User must be a normal user or a root user to delete data.')

        if not isinstance(models, list):
            models = [models]

        # Delete the resources
        for model in models:
            self._context_data.db_session.delete(model)


class UserScopedDeleter(Deleter):
    """Deleter for UserScoped models.

    This deleter should be used for UserScoped models, like Tags and APITokens.
    """

    def delete(self, models: list[T] | T) -> None:
        """Delete the UserScoped data.

        We override this method from the superclass because we have to make
        sure the `user_id` is set to the correct value first. If this field is
        set to a wrong user_id, we raise an exception.

        Args:
            models: the models to delete.
        """
        models = self._validate_user_scoped_models(models)
        super().delete(models)


class UserDeleter(Deleter):
    """Deleter for User models.

    This deleter should be used to delete Users.
    """

    def delete(self, models: list[T] | T) -> None:
        """Delete the User data.

        We override this method from the superclass because we have to make
        sure the model is a User model and that the user in the context is
        allowed to delete this user. A root user can delete ALL users, except
        his own user.

        Args:
            models: the models to delete.

        Raises:
            WrongDataManipulatorException: when the model in the instance is
                not a User.
            PermissionDeniedException: when the model is not the same model as
                set in the instance, when the model is for the current user or
                when the user not allowed to remove this User.
        """
        if self._database_model is not User:
            raise WrongDataManipulatorException(
                f'The model "{self._database_model}" is not a User')

        # Make sure the `models` are always a list
        if not isinstance(models, list):
            models = [models]

        if self._context_data.user.role == UserRole.USER:
            raise PermissionDeniedException(
                'A normal user cannot remove users')

        # Check for all models are a User model and if the `id` field is the
        # same as the current user if this user is a USER user.
        for model in models:
            if not isinstance(model, self._database_model):
                raise PermissionDeniedException(
                    f'Expected "{self._database_model}", got "{type(model)}".')

            if self._context_data.user.id == model.id:
                raise PermissionDeniedException(
                    'Cannot remove the current user.')

        super().delete(models)



"""Module with Retrievers.

This module contains the Retriever classes. These classes are used to retrieve
data from the database. The ResourceManager uses these classes.
"""

from typing import TypeVar

from my_model.user_scoped_models import (User, UserRole,  # type: ignore
                                         UserScopedModel)
from sqlalchemy.sql.elements import ColumnElement
from sqlmodel import select

from .data_manipulator import DataManipulator
from .exceptions import (BaseClassCallException, PermissionDeniedException,
                         WrongDataManipulatorException)

T = TypeVar('T')


class Retriever(DataManipulator):
    """Baseclass for Retrievers.

    The baseclass for retrievers. The sub retrievers use this class to make
    sure all retrievers have the same interface.
    """

    def get_context_filters(self) -> list[ColumnElement]:
        """Set default filters for the object.

        Method that returns the default filters for this specific getters. This
        should be used when retrieving resources so only the correct resources
        are retrieved. This method should be overridden by subclasses.

        Raises:
            BaseClassCallException: BaseClass method is used.
        """
        raise BaseClassCallException('Method not implemented in baseclass')

    def retrieve(
            self,
            flt: list[ColumnElement] | ColumnElement | None = None) -> list[T]:
        """Retrieve data.

        The method to retrieve data from the database. Can only be done by
        normal users and root users. A Service user, for instance, cannot
        use this method to retrieve data.

        Args:
            flt: a SQLalchemy filter to filter the retrieved data. Can be a
                list of filters, or a single filter.

        Raises:
            PermissionDeniedException: when this user type is not allowed to
                retrieve data this way.

        Returns:
            A list with retrieved data. If no data was found, a empty list is
            returned. If only one item is found, a list with one element is
            returned.
        """
        if self._context_data.user.role not in (UserRole.USER, UserRole.ROOT):
            raise PermissionDeniedException(
                'User must be a normal user or a root user to retrieve data.')

        # Retrieve the resources
        sql_query = select(self._database_model)

        # Filter on the context-based filters
        for filter_item in self.get_context_filters():
            sql_query = sql_query.where(filter_item)

        if isinstance(flt, ColumnElement):
            flt = [flt]

        # Add the filters from the command line
        if flt:
            for filter_item in flt:
                sql_query = sql_query.where(filter_item)

        resources = self._context_data.db_session.exec(sql_query).all()

        # Return the given resources
        return resources


class UserScopedRetriever(Retriever):
    """Retriever for UserScoped models.

    This retriever should be used for UserScoped models, like Tags and
    APITokens.
    """

    def get_context_filters(self) -> list[ColumnElement]:
        """Get default filters for the current context.

        Returns the fitlers that are relevant for the current context. In the
        case of UserScoped models, this method will return filters to make sure
        only the data for the user in the context are returned.

        If the model is not a UserScoped model, the method will raise an
        exception.

        Raises:
            WrongDataManipulatorException: when the model for the class is not
                a UserScoped model.

        Returns:
            A list with the SQLalchmey filters.
        """
        if not issubclass(self._database_model, UserScopedModel):
            raise WrongDataManipulatorException(
                f'The model "{self._database_model}" is not a UserScopedModel')
        return [self._database_model.user_id == self._context_data.user.id]


class UserRetriever(Retriever):
    """Retriever for Users.

    This retrieved should be used for User models.
    """

    def get_context_filters(self) -> list[ColumnElement]:
        """Get default filters for the current context.

        Returns the fitlers that are relevant for the current context. In this
        case, it returns no filters is the user is a ROOT user, since that user
        is allowed to see all users. If the user is a normal user, it returns
        a filter that will make sure only the own user will be selected.

        If the model is not a User model, the method will raise an exception.

        Raises:
            WrongDataManipulatorException: when the model for the class is not
                a User model.

        Returns:
            A list with the SQLalchmey filters.
        """
        if self._database_model is not User:
            raise WrongDataManipulatorException(
                f'The model "{self._database_model}" is not a User')
        if self._context_data.user.role == UserRole.USER:
            return [User.id == self._context_data.user.id]

        # Root users get no filter
        return []



"""My Data Package.

This package contains the code to create, retrieve, update and delete data from
a database containing the objects for the `my_model` package.

To use this package, you have to use the MyData class, which is located in the
`my_data` module. After creating a instance of this class, you have to
configure it with the correct database string and additional arguments. These
arguments are all used by SQLmodel (and SQLalchemy) to create the correct
engine to connect to the database.

An example:

```python
from my_data import

# Configure the database
my_data = MyData()
my_data.configure(
    # db_connection_str='sqlite:///:memory:',
    db_connection_str='sqlite:////home/dast/my.sqlite',
    database_args={
        'echo': True,
        'pool_pre_ping': True,
        'pool_recycle': True,
    }
)

# Create the engine and the test data
my_data.create_engine()
```

You can now use the package. If you need to create the tables, you can use the
`create_db_tables` method:

```python
my_data.create_db_tables()
```

If you want to create initial data, for testing or when initializing the
database, you can use the `create_init_data` tables. _Warning:_ this will
remove all tables in the database first, so use with care!

```python
my_data.create_init_data()
```

After that, a Context can be used to manipulate data in the database:

```python
with my_data.get_context(user=root_user) as c:
    all_users = c.user.retrieve()
```
"""

from .my_data import MyData  # noqa: F401

__version__ = '1.0.0'



"""Module with the DataManipulator class.

This module contains the DataManipulator class. This class is used as baseclass
for other DataManipulator classes.
"""

from typing import Generic, Type, TypeVar

from my_model.user_scoped_models import UserScopedModel  # type: ignore
from sqlalchemy.future import Engine

from .context_data import ContextData
from .exceptions import (PermissionDeniedException,
                         WrongDataManipulatorException)

T = TypeVar('T')


class DataManipulator(Generic[T]):
    """Class to manipulate database data.

    The DataManipulator class is the baseclass for other DataManipulators.
    Subclasses inherit this class to get a consistent initiator.

    Attributes:
        _database_model: the SQLmodel model used by this DataManipulator.
        _database_engine: the SQLalchemy engine to use.
        _context_data: specifies in what context to use the manipulator.
    """

    def __init__(self,
                 database_model: Type[T],
                 database_engine: Engine,
                 context_data: ContextData) -> None:
        """Set attributes for the class.

        The initiator sets the attributes for the class to the values specified
        in the arguments.

        Args:
            database_model: the SQLmodel model used by this DataManipulator.
            database_engine: the SQLalchemy engine to use.
            context_data: specifies in what context to use the manipulator.
        """
        self._database_model = database_model
        self._database_engine = database_engine
        self._context_data = context_data

    def _convert_model_to_list(self, models: list[T] | T) -> list[T]:
        """Convert a model to a list of models.

        Method to convert a model to a list of models, unless it already is a
        list.

        Args:
            models: the model(s).

        Returns:
            A list with models.
        """
        if not isinstance(models, list):
            return [models]
        return models

    def _validate_user_scoped_models(self, models: list[T] | T) -> list[T]:
        """Validate model type and user ID in user scoped models.

        Method to validate if a User Scoped model is a subclass of the
        baseclass UserScopedModel and if the `user_id` field in the data is set
        to the user in the context.

        Args:
            models: the models to check.

        Raises:
            WrongDataManipulatorException: when the model in the instance is
                not a UserScopedModel.
            PermissionDeniedException: when the model is not the same model as
                set in the instance or when the model has a user_id set that is
                different then the current user_id in the context.

        Returns:
            A list with the resources.
        """
        # Check if it is a subtype of UserScopedModel
        if not issubclass(self._database_model, UserScopedModel):
            raise WrongDataManipulatorException(
                f'The model "{self._database_model}" is not a UserScopedModel')

        # Make sure the `models` are always a list
        models = self._convert_model_to_list(models)

        # Verify the model type and if the `user_id` field is set.
        for model in models:
            if not isinstance(model, self._database_model):
                raise PermissionDeniedException(
                    f'Expected "{self._database_model}", got "{type(model)}".')

            if getattr(model, 'user_id', None) != self._context_data.user.id:
                raise PermissionDeniedException(
                    'This user is not allowed to alter this resource')

        return models

    def _add_models_to_session(self, models: list[T] | T) -> list[T]:
        """Add models to a session and commit the session.

        Method to add models a SQLalchemy session and commit the session. This
        can be used for adding or updating resources.

        Args:
            models: the models to add.

        Returns:
            The list of models.
        """
        # Make sure the `models` are always a list
        if not isinstance(models, list):
            models = [models]

        # Update the resources
        for model in models:
            self._context_data.db_session.add(model)
        return models
